#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# saml_crypto.py
#
# librairie de signature de requêtes SAML 2.0
# nécessite les librairies suivantes : 
# - python-xmlsec (http://pyxmlsec.labs.libre-entreprise.org)
# - python-M2Crypto
#
###########################################################################

import os, traceback
import base64, urllib
import libxml2, xmlsec
from M2Crypto import EVP, m2

try:
    _ = _
except NameError:
    _ = unicode

######################
# utilities functions

def load_trusted_certs(cert):
    mngr = xmlsec.KeysMngr()
    if xmlsec.cryptoAppDefaultKeysMngrInit(mngr) < 0:
        print "Error: failed to initialize keys manager."
        mngr.destroy()
        return None
    if not os.access(cert, os.R_OK):
        mngr.destroy()
        return None
    # Load trusted cert
    if mngr.certLoad(cert, xmlsec.KeyDataFormatPem,
                     xmlsec.KeyDataTypeTrusted) < 0:
        print "Error: failed to load pem certificate from \"%s\"", cert
        mngr.destroy()
        return None
    return mngr

def init_libs():
    libxml2.initParser()
    xmlsec.init()
    # xmlsec.cryptoAppInit(None)
    xmlsec.cryptoInit()

def cleanup(doc=None, dsig_ctx=None, res=False):
    if dsig_ctx is not None:
        dsig_ctx.destroy()
    if doc is not None:
        doc.freeDoc()
    xmlsec.cryptoShutdown()
    # xmlsec.cryptoAppShutdown()
    xmlsec.shutdown()
    libxml2.cleanupParser()
    return res

def dtd_log(msg, arg):
    """prevents dtd validation warnings to be logged"""
    pass

def find_key_file(cert_file):
    try:
        if os.path.isfile(os.path.splitext(cert_file)[0] + '.key'):
            key_file = os.path.splitext(cert_file)[0] + '.key'
        else:
            key_file = cert_file
    except:
        # key not found in certificate file, searching in corresponding .key file (if present)
        key_file = cert_file
    return key_file

################################################
# sign/check request methods used for signature
# of request arguments in HTTP-Redirect binding

def sign_request(cert_file, sign_method, message_type, message, relay_state=None):
    """generates signature for given SAML message using openssl_sign (used for Redirect binding)

    signature should be applied on (can be either SAMLResponse or SAMLRequest) : 
    
        SAMLResponse=resp_data&RelayState=relay_data&SigAlg=value -> encoded in location header

    resp_data -> urlencode( base64_encode( gzdeflate( request ) ))
    relay_data -> urlencode( relay_state )
    SigAlg -> urlencode( sign_method )
    """
    req_args = []
    relay_info = ""
    req_args.append((message_type, message))
    if relay_state:
        req_args.append(('RelayState', relay_state))
    req_args.append(('SigAlg', sign_method))
    req_string = urllib.urlencode(req_args)
    # load private key and sign data
    pkey_file = find_key_file(cert_file)
    k = EVP.load_key(pkey_file)
    k.sign_init()
    k.sign_update(req_string)
    signature = base64.encodestring(k.final())
    # append signature data to request args
    req_args.append(('Signature', signature))
    return req_args

def check_signed_request(cert_file, args, message_type):
    """checks that request arguments for a saml Response/Request
    correspond to the Signature argument

    args : the request arguments
    message_type : the argument which contains the saml message
    """
    req_args = []
    msg = "%s %s" % (message_type, str(_("signature could not be validated")))
    # only rsa-sha1 signature is supported
    try:
        if 'Signature' in args and args.get('SigAlg',[''])[0] == xmlsec.HrefRsaSha1:
            # fetching signature value
            signature_value = base64.decodestring(args['Signature'][0])
            # we rebuild the original string 
            if message_type in args:
                req_args.append((message_type, args[message_type][0]))
                if 'RelayState' in args:
                    req_args.append(('RelayState', args['RelayState'][0]))
                if 'SigAlg' in args:
                    req_args.append(('SigAlg', args['SigAlg'][0]))
                    # all required parameters are present
                    req_string = urllib.urlencode(req_args)
                    # load private key and sign data
                    pkey_file = find_key_file(cert_file)
                    k = EVP.load_key(pkey_file)
                    k.verify_init()
                    k.verify_update(req_string)
                    # checking that signature is valid
                    # HACK : we should use k.verify_final, but as it is broken
                    # in older m2crypto versions, we use the underlying m2 module
                    if m2.verify_final(k.ctx, signature_value, k.pkey):
                        return True, ""
                    else:
                        msg = _("invalid signature value")
    except Exception, e:
        traceback.print_exc()
        msg = str(e)
    return False, msg
        

#####################################################
# sign/check xml document methods used for signature
# of xml data in HTTP-POST binding

def sign_doc(xml_src, cert_file, reference=None, key_info=True):
    """signs an xml document
    references : list of references ids of nodes to sign, sign the whole document if None
    """
    init_libs()
    # validation with minimal DTD to permit IDs to be referenced
    doc = libxml2.parseDoc(xml_src)
    ctx = libxml2.newValidCtxt()
    ctx.setValidityErrorHandler(dtd_log, dtd_log)
    dtd = libxml2.parseDTD(None, "saml.dtd")
    # validating document (necessary for ID processing)
    doc.validateDtd(ctx, dtd)
    # Create signature template for RSA-SHA1 enveloped signature
    signNode = xmlsec.TmplSignature(doc, xmlsec.transformExclC14NId(),xmlsec.transformRsaSha1Id(), None)
    parent = doc.getRootElement()
    sibling = None
    # searching node for signNode insertion (global or in assertion)
    if reference:
        for ch_node in doc.getRootElement().get_children():
            if ch_node.name == 'Assertion':
                id_attr = ch_node.hasProp('ID')
                if id_attr and id_attr.content == reference:
                    for ass_child in ch_node.get_children():
                        if ass_child.name == 'Issuer':
                            sibling = ass_child
                    break
    else:
        for ch_node in doc.getRootElement().get_children():
            if ch_node.name == 'Issuer':
                sibling = ch_node
                break
    # Add <Signature/> node to the doc
    if sibling:
        sibling.addNextSibling(signNode)
    else:
        parent.addChild(signNode)
    # Create signature context, we don't need keys manager in this example
    dsig_ctx = xmlsec.DSigCtx()
    # Add reference
    if reference:
        refNode = signNode.addReference(xmlsec.transformSha1Id(), None, "#%s" % reference, None)
        dsig_ctx.enabledReferenceUris = xmlsec.TransformUriTypeSameDocument
        dsig_ctx.keyInfoReadCtx.retrievalMethodCtx.enabledUris = xmlsec.TransformUriTypeSameDocument
    else:
        refNode = signNode.addReference(xmlsec.transformSha1Id(), None, None, None)
    # Add enveloped transform
    refNode.addTransform(xmlsec.transformEnvelopedId())
    refNode.addTransform(xmlsec.transformExclC14NId())
    key_file = find_key_file(cert_file)
    key = xmlsec.cryptoAppKeyLoad(key_file, xmlsec.KeyDataFormatPem, None, None, None)
    # Load certificate and add to the key
    if xmlsec.cryptoAppKeyCertLoad(key, cert_file, xmlsec.KeyDataFormatPem) < 0:
        cleanup(doc, dsig_ctx)
        raise Exception("Error: failed to load pem certificate \"%s\"" % cert_file)
    # Set key name to the file name
    if key.setName(key_file) < 0:
        cleanup(doc, dsig_ctx)
        raise Exception("Error: failed to set key name for key from \"%s\"" % key_file)

    if key_info:
        # Add <dsig:KeyInfo/> and <dsig:X509Data/>
        keyInfoNode = signNode.ensureKeyInfo(None)
        X509Node = keyInfoNode.addX509Data()

    dsig_ctx.signKey = key
    # Sign the template
    if dsig_ctx.sign(signNode) < 0:
        cleanup(doc, dsig_ctx)
        raise Exception("Error: signature failed")
    # return document as a string
    data = doc.serialize()
    cleanup(doc, dsig_ctx)
    return xmlsec.HrefRsaSha1, data

def verify_doc(mngr, xml_src):
    """Verifies XML signature in xml_file.
    Returns 0 on success or a negative value if an error occurs.
    """
    status = False
    libxml2.initParser()
    doc = libxml2.parseDoc(xml_src)
    if doc is None or doc.getRootElement() is None:
        print "Error: unable to parse request"
        return cleanup(doc)
    # Find start node
    node = xmlsec.findNode(doc.getRootElement(),
                           xmlsec.NodeSignature, xmlsec.DSigNs)
    # Create signature context
    dsig_ctx = xmlsec.DSigCtx(mngr)
    if node is None:
        print "Error: signature node not found in request"
        return cleanup(doc, dsig_ctx)
    # Verify signature
    if dsig_ctx.verify(node) < 0:
        print "Error: signature verify"
        return cleanup(doc, dsig_ctx)
    # Print verification result to stdout
    if dsig_ctx.status == xmlsec.DSigStatusSucceeded:
        status = True
        print "Signature is OK"
    else:
        print "Signature is INVALID"
    # Success
    return cleanup(doc, dsig_ctx, status)

def check_signed_doc(cert_file, xml_src):
    """signs an xml document"""
    try:
        xmlsec.init()
        #xmlsec.cryptoAppInit(None)
        xmlsec.cryptoInit()
        # Create keys manager and load trusted certificates
        mngr = load_trusted_certs(cert_file)
        # Verify file
        if mngr is not None:
            res = verify_doc(mngr, xml_src)
            # Destroy keys manager
            mngr.destroy()
        # Shutdown xmlsec-crypto library
        xmlsec.cryptoShutdown()
        #xmlsec.cryptoAppShutdown()
        xmlsec.shutdown()
    except Exception, e:
        traceback.print_exc()
        return False, _("An error occured while checking signature")
    if res:
        return True, ""
    else:
        return False, _("invalid XML signature")

