#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2007
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# saml_response.py
#
# librairie de génération de requêtes SAML 2.0 spécifiques à eole-sso
# nécessite les librairies suivantes : 
# - python-saml (Copyright (C) 2006 Google Inc - http://www.apache.org/licenses/LICENSE-2.0)
# - python xmlsec (http://pyxmlsec.labs.libre-entreprise.org)
#
###########################################################################

from saml2 import saml, samlp
from saml_utils import format_timestamp
from saml_crypto import sign_doc
import time
import xmlsec
from eolesso.util import gen_random_id
from config import AUTH_SERVER_ADDR, ip_locale

# missing definitions in saml2 library
NAMEID_FORMAT_TRANSIENT = 'urn:oasis:names:tc:SAML:2.0:nameid-format:transient'
NAMEID_FORMAT_EMAIL = 'urn:oasis:names:tc:SAML:2.0:nameid-format:email'
URN_PROTECTED_PASSWORD = 'urn:oasis:names:tc:SAML:2.0:ac:classes:PasswordProtectedTransport'
URN_PREVIOUS_SESSION = 'urn:oasis:names:tc:SAML:2.0:ac:classes:PreviousSession'
NAME_FORMAT_BASIC = 'urn:oasis:names:tc:SAML:2.0:attrname-format:basic'

#STATEMENT_TIMEOUT = 7*24*3600
STATEMENT_TIMEOUT = 600
TIME_ADJUST = -300

def gen_assertion(user_id, attributes, creation_date, assertion_consumer, assertion_id, issuer, sp_ident, session_index, from_credentials, cert_file):
    # creating assertion
    assertion = saml.Assertion(version="2.0", id=assertion_id, issue_instant = format_timestamp(creation_date + TIME_ADJUST))
    assertion.issuer = saml.Issuer(text=issuer)
    # assertion subject
    # name_id_format = "urn:oasis:names:tc:SAML:2.0:nameid-format:transient" ??
    #                  "urn:oasis:names:tc:SAML:1.1:nameid-format:emailAddress"  (cf SimpleSaml)
    # le format peut être défini dans les metadata (priorité : sp -> idp -> default)
    nid = saml.NameID(format=NAMEID_FORMAT_TRANSIENT,name_qualifier=issuer, sp_name_qualifier=sp_ident, text=user_id)
    subj = saml.Subject(name_id=nid)
    # subject confirmation
    subj_conf = saml.SubjectConfirmation(method=saml.SUBJECT_CONFIRMATION_METHOD_BEARER)
    conf_data = saml.SubjectConfirmationData(not_on_or_after=format_timestamp(creation_date + STATEMENT_TIMEOUT), recipient=assertion_consumer)
    subj_conf.subject_confirmation_data = conf_data
    subj.subject_confirmation = subj_conf
    assertion.subject = subj
    assertion.conditions = gen_conditions(assertion_consumer, creation_date)
    assertion.authn_statement = gen_statement(creation_date, session_index, from_credentials)
    # attributes statements
    assertion.attribute_statement = gen_attributes(attributes)
    return assertion

def gen_attributes(attributes):
    attr_statement = saml.AttributeStatement()
    for attr_name, values in attributes.items():
        if type(values) != list:
            values = [values]
        for value in values:
            if value != "":
                # si besoin : encoder en base64 et 
                attr = saml.Attribute(name=attr_name, name_format=NAME_FORMAT_BASIC)
                attr.attribute_value = saml.AttributeValue(text=value,extension_attributes={'xsi:type':"xs:string"})
                attr_statement.attribute.append(attr)
    return attr_statement

def gen_statement(creation_date, session_index, from_credentials):
    if from_credentials:
        class_ref=URN_PROTECTED_PASSWORD
    else:
        class_ref=URN_PREVIOUS_SESSION
    # XXX FIXME : Forced to PROTECTED_PASSWORD for fim even if authenticated from a previous session
    class_ref=URN_PROTECTED_PASSWORD
    # subject locality
    subj_loc = saml.SubjectLocality(address=ip_locale, dns_name=AUTH_SERVER_ADDR)
    # auth context
    ctx = saml.AuthnContext(authn_context_class_ref=saml.AuthnContextClassRef(text=class_ref))
    st = saml.AuthnStatement(authn_instant=format_timestamp(creation_date + TIME_ADJUST),
                             session_index="%s" % str(session_index),
                             subject_locality = subj_loc,
                             authn_context=ctx)
    return st


def gen_conditions(assertion_consumer, creation_date):
    audience = saml.Audience(text=assertion_consumer)
    audience_restr = saml.AudienceRestriction(audience=audience)
    return saml.Conditions(not_before=format_timestamp(creation_date), not_on_or_after=format_timestamp(creation_date + STATEMENT_TIMEOUT), audience_restriction=audience_restr)
    # return saml.Conditions(not_on_or_after=format_timestamp(creation_date + STATEMENT_TIMEOUT), audience_restriction=audience_restr)

def gen_response(response_id, authenticated, session_index, from_credentials, user_id, attributes, saml_request, assertion_consumer, issuer, sp_ident, server_url, cert_file):
    response = samlp.Response(extension_attributes = {'xmlns:xs':"http://www.w3.org/2001/XMLSchema", 
                                                      'xmlns:saml':"urn:oasis:names:tc:SAML:2.0:assertion",
                                                      'xmlns:xsi':"http://www.w3.org/2001/XMLSchema-instance",
                                                      'xmlns:ds':"http://www.w3.org/2000/09/xmldsig#"}
                             )
    response.id = response_id
    if saml_request:
        response.in_response_to = "request id"
    response.version = saml.V2
    creation_date = time.time()
    response.issue_instant = format_timestamp(creation_date + TIME_ADJUST)
    response.destination = assertion_consumer
    response.consent = saml.CONSENT_UNSPECIFIED
    response.issuer = saml.Issuer(text=issuer)
    response.status = samlp.Status()

    if not authenticated:
        response.status.status_code = samlp.StatusCode(samlp.STATUS_AUTHN_FAILED)
        response.status.status_message = samlp.StatusMessage()
        response.status.status_detail = samlp.StatusDetail()
        assertion_id = None
    else:
        response.status.status_code = samlp.StatusCode(samlp.STATUS_SUCCESS)
        response.status.status_message = samlp.StatusMessage()
        response.status.status_detail = samlp.StatusDetail()
        assertion_id = gen_random_id('_')
        response.assertion.append(gen_assertion(user_id, attributes, creation_date, assertion_consumer, assertion_id, issuer, sp_ident, session_index, from_credentials, cert_file))
    xml_doc = response.ToString()
    # sign and return the document
    return sign_doc(xml_doc, cert_file, assertion_id)

def gen_logout_response(request_id, response_id, issuer, destination, cert_file, logout_status, sign=True):
    logout_msg = {samlp.STATUS_SUCCESS: _('Single logout success'),
                  samlp.STATUS_PARTIAL_LOGOUT: _('Partial single logout success')
                 }
    response = samlp.LogoutResponse(in_response_to=request_id, id=response_id)
    response.version = saml.V2
    creation_date = time.time()
    response.issue_instant = format_timestamp(creation_date + TIME_ADJUST)
    response.not_on_or_after = format_timestamp(creation_date + STATEMENT_TIMEOUT)
    response.issuer = saml.Issuer(text=issuer)
    response.destination = destination
    response.status = samlp.Status()
    response.status.status_code = samlp.StatusCode(logout_status)
    response.status.status_message = samlp.StatusMessage(text=str(logout_msg[logout_status]))
    response.status.status_detail = samlp.StatusDetail()
    xml_doc = response.ToString()
    if not sign:
        # include assertion in a string and sign the string
        return xmlsec.HrefRsaSha1, xml_doc
    else:
        # sign within xml
        return sign_doc(xml_doc, cert_file)

def gen_logout_request(request_id, session_index, user_id, issuer, sp_ident, destination, cert_file, sign=True):
    request = samlp.LogoutRequest(id=request_id)
    request.version = saml.V2
    creation_date = time.time()
    request.issue_instant = format_timestamp(creation_date + TIME_ADJUST)
    request.not_on_or_after = format_timestamp(creation_date + STATEMENT_TIMEOUT)
    request.issuer = saml.Issuer(text=issuer)
    request.destination = destination
    request.session_index = samlp.SessionIndex(text=session_index)
    request.name_id = saml.NameID(format=NAMEID_FORMAT_TRANSIENT,text=user_id, name_qualifier=issuer, sp_name_qualifier=sp_ident)
    request.consent = saml.CONSENT_UNSPECIFIED
    request.issuer = saml.Issuer(text=issuer)
    xml_doc = request.ToString()
    if not sign:
        # include assertion in a string and sign the string
        return xmlsec.HrefRsaSha1, xml_doc
    else:
        # sign within xml
        return sign_doc(xml_doc, cert_file)
