#! /usr/bin/env python
# -*- coding: utf-8 -*-

###########################################################################
#
# Eole NG - 2008
# Copyright Pole de Competence Eole  (Ministere Education - Academie Dijon)
# Licence CeCill  cf /root/LicenceEole.txt
# eole@ac-dijon.fr
#
# saml_utils.py
#
# utilitaires pour la gestion des requêtes SAML et des données metadata
#
###########################################################################

import os, socket, datetime, time

# from OpenSSL import crypto
from M2Crypto import X509
from xml.etree import ElementTree
import base64, re, zlib, urllib, StringIO
from config import METADATA_DIR, IDP_IDENTITY
from saml2 import samlp
from twisted.python import log


try:
    _ = _
except NameError:
    _ = unicode

# exceptions

class Redirect(Exception):
    def __init__(self, location):
        super(Redirect, self).__init__(location)
        self.location = location

class InternalError(Exception):

    def __init__(self, reason):
        self.reason = reason
        self.code = 'INTERNAL_ERROR'
        super(InternalError, self).__init__(reason)

    def as_xml(self):
        return """%s:%s""" % (self.code, escape(self.reason))

# outils de formattage des dates

def date_from_string(date_string):
    "converts a date in SAML Format to a time object"
    return datetime.datetime.strptime(date_string, "%Y-%m-%dT%H:%M:%SZ")

def format_timestamp(ts=None):
    if ts is None:
        ts = time.time()
    date = datetime.datetime.utcfromtimestamp(ts).isoformat()
    date = date[:date.rindex('.')]
    return "%sZ" % date

def check_date(date_string, not_before = None):
    now = datetime.datetime.utcnow()
    date = date_from_string(date_string)
    if date > now:
        if (not_before is None) or (now > not_before):
            return True
    return False

# outils de decodage/encodage des requetes saml (utilisation de urlencode, base64 et gz deflate/inflate)
def htc(m):
    return chr(int(m.group(1),16)) 

def urldecode(url):
    rex=re.compile('%([0-9a-hA-H][0-9a-hA-H])',re.M)
    return rex.sub(htc,url)

def decode_request(request, compress=False):
    if compress:
        request = urldecode(request)
    request = base64.decodestring(request)
    if compress:
        # MAX_WBTS mandatory to be compatible with PHP gzinflate/deflate methods
        request = zlib.decompress(request, -zlib.MAX_WBITS)
    return request

def encode_request(request, compress=False):
    if compress:
        # compressing request with a zlib compressor object (zlib.compress can not user MAX_WBITS option)
        fp = StringIO.StringIO(request)
        cmpr = zlib.compressobj(zlib.Z_DEFAULT_COMPRESSION, zlib.DEFLATED, -zlib.MAX_WBITS)
        while 1:
            request = fp.read(1024 * 8)
            if not request:
                break
            request = cmpr.compress(request)
        request = cmpr.flush()
    request = base64.encodestring(request).replace('\n','').strip()
    return request

# outils de gestion des fichiers metadata
def find_tag(node, tag):
    md_tag = "{urn:oasis:names:tc:SAML:2.0:metadata}"
    nodes = node.findall(".//%s%s" % (md_tag, tag))
    if nodes == []:
        return node.findall(".//%s" % tag)
    else:
        return nodes

def get_metadata(identifier):
    """return an xml dom describing metadata for a given entity"""
    data = {}
    meta_file = os.path.join(METADATA_DIR, "%s.xml" % identifier)
    if not os.path.isfile(meta_file):
        raise InternalError(_('No metada found for identity %s') % identifier)
    doc = ElementTree.parse(meta_file)
    # id de l'entité
    for e in find_tag(doc, 'EntityDescriptor'):
        if 'entityID' in e.attrib: data['entityID'] = e.get('entityID')
    if not 'entityID' in data:
        if 'entityID' in doc.getroot().attrib: data['entityID'] = doc.getroot().get('entityID')
    # informations générales
    for e in find_tag(doc, 'SPSSODescriptor'):
        if 'WantAuthRequestsSigned' in e.attrib: data['WantAuthRequestsSigned'] = e.get('WantAuthRequestsSigned')
        if 'AuthnRequestsSigned' in e.attrib: data['AuthnRequestsSigned'] = e.get('AuthnRequestsSigned')
        if 'WantAssertionsSigned' in e.attrib: data['WantAssertionsSigned'] = e.get('WantAssertionsSigned')
        if 'protocolSupportEnumeration' in e.attrib: data['protocolSupportEnumeration'] = e.get('protocolSupportEnumeration')
    # informations éventuelles sur le nameID
    for e in find_tag(doc, 'NameIDFormat'):
        nameids = data.get('NameIDFormat',[])
        nameids.append(e.text)
        data['NameIDFormat'] = nameids
    # recherche de la description des urls de service (endpoints)
    for endpoint in ('AssertionConsumerService', 'SingleLogoutService'):
        for e in find_tag(doc, endpoint):
            resp_loc = ""
            is_default = False
            index = None
            if e.get('isDefault') == 'true':
                is_default = True
            if 'index' in e.attrib:
                try:
                    index = int(e.get('index'))
                except ValueError:
                    index = None
            if 'ResponseLocation' in e.attrib:
                # responses have to be sent on a different url
                resp_loc = e.get('ResponseLocation')
            if 'Location' in e.attrib:
                data_endp = data.get(endpoint,[])
                data_endp.append((e.get('Location'), e.get('Binding'), resp_loc, is_default, index))
                data[endpoint] = data_endp
    return data

def gen_metadata(server_url, cert_file):
    """generates metadata xml for this id provider"""
    # données dynamiques XXX FIXME
    identity = IDP_IDENTITY
    logout_endpoint = "%s/saml/logout" % server_url
    sso_endpoint = "%s/saml" % server_url
    cert = X509.load_cert(cert_file)
    # cert = crypto.load_certificate(crypto.FILETYPE_PEM, open(cert_file).read())
    cert_data = []
    # for line in crypto.dump_certificate(crypto.FILETYPE_PEM, cert).split('\n'):
    for line in cert.as_pem().split('\n'):
        if '-----BEGIN CERTIFICATE-----' not in line and '-----END CERTIFICATE-----' not in line:
            cert_data.append(line)
    cert_data = '\n'.join(cert_data).strip()
    # chargement du template des metadata
    md_tmpl = open('idp_metadata.tmpl').read().strip()
    data = md_tmpl % (identity, cert_data, logout_endpoint, logout_endpoint, sso_endpoint, sso_endpoint)
    return data

def get_endpoint(sp_meta, service, allowed_bindings = [samlp.BINDING_HTTP_POST, samlp.BINDING_HTTP_REDIRECT], index=None):
    service_url = binding = response_url = ""
    for endp_data in sp_meta.get(service, [("","","",False,None)]):
        service, bind, response, isdefault, ind = endp_data
        if index and ind != index:
            # index précis demandé, on passe au suivant
            continue
        # FIXME TODO : verify assertion consumer url certificate ? (agriates)
        if not bind in allowed_bindings:
            continue
        # si cet endpoint a le bon index ou est l'endpoint par défaut (ou si on n'en a pas encore trouvé), on le sélectionne
        if isdefault == True or service_url == "" or index:
            service_url = service
            binding = bind
            response_url = response
    if service_url == "":
        raise Exception('%s : %s' % ( sp_meta.get('entityId',_('service provider')), _('no usable endpoint for %s') % service))
    if not binding in allowed_bindings:
        raise Exception('%s : %s' % ( sp_meta.get('entityId',_('service provider')), _('no suppported binding for %s') % service))
    return service_url, binding, response_url

